#include <resip/stack/SipStack.hxx>
#include <resip/dum/DialogUsageManager.hxx>
#include <resip/dum/ClientSubscription.hxx>
#include <resip/dum/SubscriptionHandler.hxx>
#include <resip/dum/MasterProfile.hxx>
#include <rutil/ThreadIf.hxx>
#include <rutil/Log.hxx>
#include <rutil/Logger.hxx>

#define RESIPROCATE_SUBSYSTEM Subsystem::TEST

using namespace resip;

class ServerHandler : public ServerSubscriptionHandler
{
   public:
   
      virtual void onNewSubscription(ServerSubscriptionHandle ssh, const SipMessage&)
      {
         mSsh = ssh;

         ssh->send( ssh->accept() );

         ssh->setSubscriptionState(Active);

         resip::SharedPtr<SipMessage> pNotify  = ssh->neutralNotify();
         ssh->send(pNotify);
      }

      virtual void onTerminated(ServerSubscriptionHandle)
      {
         std::cout << "Server::onTerminated()" << std::endl;
      }

      virtual void onRefresh(resip::ServerSubscriptionHandle ssh, const SipMessage&)
      {
         ssh->send( ssh->accept() );

         resip::SharedPtr<SipMessage> pNotify = ssh->neutralNotify();
         ssh->send(pNotify);
      }

      resip::ServerSubscriptionHandle mSsh;
};

class ClientHandler : public ClientSubscriptionHandler
{
   public:
  
      ClientHandler()
      {
         mTerminated = true;
      }

      virtual void onUpdatePending(ClientSubscriptionHandle csh, const SipMessage&, bool)
      {
         csh->acceptUpdate();
      }

      virtual void onUpdateActive(ClientSubscriptionHandle csh, const SipMessage&, bool)
      {
         csh->acceptUpdate();
      }

      virtual void onUpdateExtension(ClientSubscriptionHandle csh, const SipMessage&, bool)
      {
         csh->acceptUpdate();
      }

      virtual int onRequestRetry(ClientSubscriptionHandle, int, const SipMessage&)
      {
         return -1;
      }

      virtual void onTerminated(ClientSubscriptionHandle, const SipMessage*)
      {
         std::cout << "Client::onTerminated()" << std::endl;
         mTerminated = true;
      }

      virtual void onNewSubscription(ClientSubscriptionHandle csh, const SipMessage&)
      {
         mCsh = csh;
         mTerminated = false;
      }
   
      resip::ClientSubscriptionHandle mCsh;

      bool mTerminated;

};

class Runner : public ThreadIf
{
   public:

      Runner(SipStack& stack, DialogUsageManager& dum) : mStack(stack), mDum(dum)
      {
      }

      virtual ~Runner()
      {
      }

   protected:

      virtual void thread()
      {
         while(!isShutdown())
         {
            FdSet set;
            mStack.buildFdSet(set);

            set.selectMilliSeconds(50);

            mStack.process(set);

            while(mDum.process());
         }
      }

      SipStack& mStack;
      DialogUsageManager& mDum;

};

int main(void)
{
   Log::initialize(Log::Cout, Log::Debug, "test");

   ClientHandler clientHandler;
   ServerHandler serverHandler;

   // The stack and dum instances for the server
   SipStack serverStack;
   serverStack.addTransport(UDP, 19826);
   DialogUsageManager serverDum(serverStack);
   Runner serverRunner(serverStack, serverDum);

   SharedPtr<MasterProfile> serverProfile(new MasterProfile());
   serverProfile->clearSupportedMethods();
   serverProfile->addSupportedMethod(NOTIFY);
   serverProfile->addSupportedMethod(SUBSCRIBE);

   serverDum.setMasterProfile(serverProfile);
   serverDum.addServerSubscriptionHandler("test", &serverHandler);

   // The stack and dum instances for the client
   SipStack clientStack;
   clientStack.addTransport(UDP, 19836);
   DialogUsageManager clientDum(clientStack);
   Runner clientRunner(clientStack, clientDum);
   
   SharedPtr<MasterProfile> clientProfile(new MasterProfile());
   clientProfile->clearSupportedMethods();
   clientProfile->addSupportedMethod(NOTIFY);
   clientProfile->addSupportedMethod(SUBSCRIBE);

   clientDum.setMasterProfile(clientProfile);
   clientDum.addClientSubscriptionHandler("test", &clientHandler);

   serverRunner.run();
   clientRunner.run();
   
   assert(clientHandler.mTerminated == true);

   // Create a subscription from client to server
   SharedPtr<SipMessage> subscribeMsg = clientDum.makeSubscription(NameAddr("sip:127.0.0.1:19826"), "test", 20);

   clientDum.send(subscribeMsg);

   // Give it some time to get processed...
   sleep(2);

   assert(clientHandler.mTerminated == false);
  
   SipMessage other;

   // Send one NOTIFY and cache it. We'll be resending it again later...
   resip::SharedPtr<SipMessage> pNotify = serverHandler.mSsh->neutralNotify();
   other = *pNotify;

   serverDum.send(pNotify);
   
   // Give it some time to get processed...
   sleep(2);

   // Now have the client end the subscription. It should send an unSUBSCRIBE
   // and expect a NOTIFY/terminated

   // However, immediately after calling end(), send an out of order (active)
   // NOTIFY based on the cached one

   // I'm hoping this happens before the client's thread get to receive the
   // NOTIFY/terminated... It seems to work on my test machine.

   clientHandler.mCsh->end();

   // Send the cached NOTIFY
   serverStack.send(other);

   // Give it some time to get processed...
   sleep(2);
   
   // The bug is that the subscription never gets terminated at the client, so
   // this assertion fails
   assert(clientHandler.mTerminated == true);
   
   serverRunner.shutdown();
   clientRunner.shutdown();
   
   serverRunner.join();
   clientRunner.join();

   return 0;
}

